from aerosandbox import ExplicitAnalysis
from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
from aerosandbox.aerodynamics.aero_3D.singularities.uniform_strength_horseshoe_singularities import \
    calculate_induced_velocity_horseshoe
from aerosandbox.aerodynamics.aero_3D.singularities.point_source import calculate_induced_velocity_point_source
from typing import Dict, Any, List, Callable, Union
from aerosandbox.aerodynamics.aero_3D.aero_buildup import AeroBuildup
from dataclasses import dataclass


### Define some helper functions that take a vector and make it a Nx1 or 1xN, respectively.
# Useful for broadcasting with matrices later.
def tall(array):
    return np.reshape(array, (-1, 1))


def wide(array):
    return np.reshape(array, (1, -1))


class LiftingLine(ExplicitAnalysis):
    """
    An implicit aerodynamics analysis based on lifting line theory, with modifications for nonzero sweep
    and dihedral + multiple wings.

    Nonlinear, and includes viscous effects based on 2D data.

    Usage example:
        >>> analysis = asb.LiftingLine(
        >>>     airplane=my_airplane,
        >>>     op_point=asb.OperatingPoint(
        >>>         velocity=100, # m/s
        >>>         alpha=5, # deg
        >>>         beta=4, # deg
        >>>         p=0.01, # rad/sec
        >>>         q=0.02, # rad/sec
        >>>         r=0.03, # rad/sec
        >>>     )
        >>> )
        >>> outputs = analysis.run()
    """

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 xyz_ref: List[float] = None,
                 model_size: str = "medium",
                 run_symmetric_if_possible: bool = False,
                 verbose: bool = False,
                 spanwise_resolution: int = 4,
                 spanwise_spacing_function: Callable[[float, float, float], np.ndarray] = np.cosspace,
                 vortex_core_radius: float = 1e-8,
                 align_trailing_vortices_with_wind: bool = False,
                 ):
        """
        Initializes and conducts a LiftingLine analysis.

        Args:

            airplane: An Airplane object that you want to analyze.

            op_point: The OperatingPoint that you want to analyze the Airplane at.

            run_symmetric_if_possible: If this flag is True and the problem fomulation is XZ-symmetric, the solver will
            attempt to exploit the symmetry. This results in roughly half the number of governing equations.

            opti: An asb.Opti environment.

                If provided, adds the governing equations to that instance. Does not solve the equations (you need to
                call `sol = opti.solve()` to do that).

                If not provided, creates and solves the governing equations in a new instance.

        """
        super().__init__()

        ### Set defaults
        if xyz_ref is None:
            xyz_ref = airplane.xyz_ref

        ### Initialize
        self.airplane = airplane
        self.op_point = op_point
        self.xyz_ref = xyz_ref
        self.model_size = model_size
        self.verbose = verbose
        self.spanwise_resolution = spanwise_resolution
        self.spanwise_spacing_function = spanwise_spacing_function
        self.vortex_core_radius = vortex_core_radius
        self.align_trailing_vortices_with_wind = align_trailing_vortices_with_wind

        ### Determine whether you should run the problem as symmetric
        self.run_symmetric = False
        if run_symmetric_if_possible:
            raise NotImplementedError("LiftingLine with symmetry detection not yet implemented!")
            # try:
            #     self.run_symmetric = (  # Satisfies assumptions
            #             self.op_point.beta == 0 and
            #             self.op_point.p == 0 and
            #             self.op_point.r == 0 and
            #             self.airplane.is_entirely_symmetric()
            #     )
            # except RuntimeError:  # Required because beta, p, r, etc. may be non-numeric (e.g. opti variables)
            #     pass

    def __repr__(self):
        return self.__class__.__name__ + "(\n\t" + "\n\t".join([
            f"airplane={self.airplane}",
            f"op_point={self.op_point}",
            f"xyz_ref={self.xyz_ref}",
        ]) + "\n)"

    @dataclass
    class AeroComponentResults:
        s_ref: float  # Reference area [m^2]
        c_ref: float  # Reference chord [m]
        b_ref: float  # Reference span [m]
        op_point: OperatingPoint
        F_g: List[Union[float, np.ndarray]]  # An [x, y, z] list of forces in geometry axes [N]
        M_g: List[Union[float, np.ndarray]]  # An [x, y, z] list of moments about geometry axes [Nm]

        def __repr__(self):
            F_w = self.F_w
            M_b = self.M_b
            return self.__class__.__name__ + "(\n\t" + "\n\t".join([
                f"L={-F_w[2]},",
                f"Y={F_w[1]},",
                f"D={-F_w[0]},",
                f"l_b={M_b[0]},",
                f"m_b={M_b[1]},",
                f"n_b={M_b[2]},",
            ]) + "\n)"

        @property
        def F_b(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of forces in body axes [N]
            """
            return self.op_point.convert_axes(*self.F_g, from_axes="geometry", to_axes="body")

        @property
        def F_w(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of forces in wind axes [N]
            """
            return self.op_point.convert_axes(*self.F_g, from_axes="geometry", to_axes="wind")

        @property
        def M_b(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of moments about body axes [Nm]
            """
            return self.op_point.convert_axes(*self.M_g, from_axes="geometry", to_axes="body")

        @property
        def M_w(self) -> List[Union[float, np.ndarray]]:
            """
            An [x, y, z] list of moments about wind axes [Nm]
            """
            return self.op_point.convert_axes(*self.M_g, from_axes="geometry", to_axes="wind")

        @property
        def L(self) -> Union[float, np.ndarray]:
            """
            The lift force [N]. Definitionally, this is in wind axes.
            """
            return -self.F_w[2]

        @property
        def Y(self) -> Union[float, np.ndarray]:
            """
            The side force [N]. Definitionally, this is in wind axes.
            """
            return self.F_w[1]

        @property
        def D(self) -> Union[float, np.ndarray]:
            """
            The drag force [N]. Definitionally, this is in wind axes.
            """
            return -self.F_w[0]

        @property
        def l_b(self) -> Union[float, np.ndarray]:
            """
            The rolling moment [Nm] in body axes. Positive is roll-right.
            """
            return self.M_b[0]

        @property
        def m_b(self) -> Union[float, np.ndarray]:
            """
            The pitching moment [Nm] in body axes. Positive is nose-up.
            """
            return self.M_b[1]

        @property
        def n_b(self) -> Union[float, np.ndarray]:
            """
            The yawing moment [Nm] in body axes. Positive is nose-right.
            """
            return self.M_b[2]

    def run(self) -> Dict:
        """
        Computes the aerodynamic forces.

        Returns a dictionary with keys:

            - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
            - 'F_b' : an [x, y, z] list of forces in body axes [N]
            - 'F_w' : an [x, y, z] list of forces in wind axes [N]
            - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
            - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
            - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
            - 'L' : the lift force [N]. Definitionally, this is in wind axes.
            - 'Y' : the side force [N]. This is in wind axes.
            - 'D' : the drag force [N]. Definitionally, this is in wind axes.
            - 'l_b', the rolling moment, in body axes [Nm]. Positive is roll-right.
            - 'm_b', the pitching moment, in body axes [Nm]. Positive is pitch-up.
            - 'n_b', the yawing moment, in body axes [Nm]. Positive is nose-right.
            - 'CL', the lift coefficient [-]. Definitionally, this is in wind axes.
            - 'CY', the sideforce coefficient [-]. This is in wind axes.
            - 'CD', the drag coefficient [-]. Definitionally, this is in wind axes.
            - 'Cl', the rolling coefficient [-], in body axes
            - 'Cm', the pitching coefficient [-], in body axes
            - 'Cn', the yawing coefficient [-], in body axes

        Nondimensional values are nondimensionalized using reference values in the LiftingLine.airplane object.

        Data types:
            - The "L", "Y", "D", "l_b", "m_b", "n_b", "CL", "CY", "CD", "Cl", "Cm", and "Cn" keys are:

                - floats if the OperatingPoint object is not vectorized (i.e., if all attributes of OperatingPoint
                are floats, not arrays).

                - arrays if the OperatingPoint object is vectorized (i.e., if any attribute of OperatingPoint is an
                array).

            - The "F_g", "F_b", "F_w", "M_g", "M_b", and "M_w" keys are always lists, which will contain either
            floats or arrays, again depending on whether the OperatingPoint object is vectorized or not.
        """
        aerobuildup = AeroBuildup(
            airplane=self.airplane,
            op_point=self.op_point,
            xyz_ref=self.xyz_ref,
        )

        wing_aero = self.wing_aerodynamics()

        fuselage_aero_components = [
            aerobuildup.fuselage_aerodynamics(
                fuselage=fuse,
                include_induced_drag=True,
            )
            for fuse in self.airplane.fuselages
        ]

        aero_components = [wing_aero] + fuselage_aero_components

        ### Sum up the forces
        F_g_total = [
            sum([comp.F_g[i] for comp in aero_components])
            for i in range(3)
        ]
        M_g_total = [
            sum([comp.M_g[i] for comp in aero_components])
            for i in range(3)
        ]

        ##### Start to assemble the output
        output = {
            "F_g": F_g_total,
            "M_g": M_g_total,
        }

        ##### Add in other metrics
        output["F_b"] = self.op_point.convert_axes(
            *F_g_total,
            from_axes="geometry",
            to_axes="body"
        )
        output["F_w"] = self.op_point.convert_axes(
            *F_g_total,
            from_axes="geometry",
            to_axes="wind"
        )
        output["M_b"] = self.op_point.convert_axes(
            *M_g_total,
            from_axes="geometry",
            to_axes="body"
        )
        output["M_w"] = self.op_point.convert_axes(
            *M_g_total,
            from_axes="geometry",
            to_axes="wind"
        )

        output["L"] = -output["F_w"][2]
        output["Y"] = output["F_w"][1]
        output["D"] = -output["F_w"][0]
        output["l_b"] = output["M_b"][0]
        output["m_b"] = output["M_b"][1]
        output["n_b"] = output["M_b"][2]

        ##### Compute dimensionalization factor
        qS = self.op_point.dynamic_pressure() * self.airplane.s_ref
        c = self.airplane.c_ref
        b = self.airplane.b_ref

        ##### Add nondimensional forces, and nondimensional quantities.
        output["CL"] = output["L"] / qS
        output["CY"] = output["Y"] / qS
        output["CD"] = output["D"] / qS
        output["Cl"] = output["l_b"] / qS / b
        output["Cm"] = output["m_b"] / qS / c
        output["Cn"] = output["n_b"] / qS / b

        ##### Add the component aerodynamics, for reference
        output["wing_aero"] = wing_aero
        output["fuselage_aero_components"] = fuselage_aero_components

        # ##### Add the drag breakdown
        # output["D_profile"] = sum([
        #     comp.D for comp in aero_components
        # ])
        # output["D_induced"] = D_induced

        return output

    def run_with_stability_derivatives(self,
                                       alpha=True,
                                       beta=True,
                                       p=True,
                                       q=True,
                                       r=True,
                                       ) -> Dict[str, Union[Union[float, np.ndarray], List[Union[float, np.ndarray]]]]:
        """
        Computes the aerodynamic forces and moments on the airplane, and the stability derivatives.

        Arguments essentially determine which stability derivatives are computed. If a stability derivative is not
        needed, leaving it False will speed up the computation.

        Args:

            - alpha (bool): If True, compute the stability derivatives with respect to the angle of attack (alpha).
            - beta (bool): If True, compute the stability derivatives with respect to the sideslip angle (beta).
            - p (bool): If True, compute the stability derivatives with respect to the body-axis roll rate (p).
            - q (bool): If True, compute the stability derivatives with respect to the body-axis pitch rate (q).
            - r (bool): If True, compute the stability derivatives with respect to the body-axis yaw rate (r).

        Returns: a dictionary with keys:

            - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
            - 'F_b' : an [x, y, z] list of forces in body axes [N]
            - 'F_w' : an [x, y, z] list of forces in wind axes [N]
            - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
            - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
            - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
            - 'L'   : the lift force [N]. Definitionally, this is in wind axes.
            - 'Y'   : the side force [N]. This is in wind axes.
            - 'D'   : the drag force [N]. Definitionally, this is in wind axes.
            - 'l_b' : the rolling moment, in body axes [Nm]. Positive is roll-right.
            - 'm_b' : the pitching moment, in body axes [Nm]. Positive is pitch-up.
            - 'n_b' : the yawing moment, in body axes [Nm]. Positive is nose-right.
            - 'CL'  : the lift coefficient [-]. Definitionally, this is in wind axes.
            - 'CY'  : the sideforce coefficient [-]. This is in wind axes.
            - 'CD'  : the drag coefficient [-]. Definitionally, this is in wind axes.
            - 'Cl'  : the rolling coefficient [-], in body axes
            - 'Cm'  : the pitching coefficient [-], in body axes
            - 'Cn'  : the yawing coefficient [-], in body axes

            Along with additional keys, depending on the value of the `alpha`, `beta`, `p`, `q`, and `r` arguments. For
            example, if `alpha=True`, then the following additional keys will be present:

                - 'CLa' : the lift coefficient derivative with respect to alpha [1/rad]
                - 'CDa' : the drag coefficient derivative with respect to alpha [1/rad]
                - 'CYa' : the sideforce coefficient derivative with respect to alpha [1/rad]
                - 'Cla' : the rolling moment coefficient derivative with respect to alpha [1/rad]
                - 'Cma' : the pitching moment coefficient derivative with respect to alpha [1/rad]
                - 'Cna' : the yawing moment coefficient derivative with respect to alpha [1/rad]
                - 'x_np': the neutral point location in the x direction [m]

            Nondimensional values are nondimensionalized using reference values in the AeroBuildup.airplane object.

            Data types:
                - The "L", "Y", "D", "l_b", "m_b", "n_b", "CL", "CY", "CD", "Cl", "Cm", and "Cn" keys are:

                    - floats if the OperatingPoint object is not vectorized (i.e., if all attributes of OperatingPoint
                    are floats, not arrays).

                    - arrays if the OperatingPoint object is vectorized (i.e., if any attribute of OperatingPoint is an
                    array).

                - The "F_g", "F_b", "F_w", "M_g", "M_b", and "M_w" keys are always lists, which will contain either
                floats or arrays, again depending on whether the OperatingPoint object is vectorized or not.

        """
        do_analysis: Dict[str, bool] = {
            "alpha": alpha,
            "beta" : beta,
            "p"    : p,
            "q"    : q,
            "r"    : r,
        }

        abbreviations: Dict[str, str] = {
            "alpha": "a",
            "beta" : "b",
            "p"    : "p",
            "q"    : "q",
            "r"    : "r",
        }
        finite_difference_amounts: Dict[str, float] = {
            "alpha": 0.001,
            "beta" : 0.001,
            "p"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.b_ref,
            "q"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.c_ref,
            "r"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.b_ref,
        }
        scaling_factors: Dict[str, float] = {
            "alpha": np.degrees(1),
            "beta" : np.degrees(1),
            "p"    : (2 * self.op_point.velocity) / self.airplane.b_ref,
            "q"    : (2 * self.op_point.velocity) / self.airplane.c_ref,
            "r"    : (2 * self.op_point.velocity) / self.airplane.b_ref,
        }

        original_op_point = self.op_point

        # Compute the point analysis, which returns a dictionary that we will later add key:value pairs to.
        run_base = self.run()

        # Note for the loops below: here, "derivative numerator" and "... denominator" refer to the quantity being
        # differentiated and the variable of differentiation, respectively. In other words, in the expression df/dx,
        # the "numerator" is f, and the "denominator" is x. I realize that this would make a mathematician cry (as a
        # partial derivative is not a fraction), but the reality is that there seems to be no commonly-accepted name
        # for these terms. (Curiously, this contrasts with integration, where there is an "integrand" and a "variable
        # of integration".)

        for d in do_analysis.keys():
            if not do_analysis[d]:  # Basically, if the parameter from the function input is not True,
                continue  # Skip this run.
                # This way, you can (optionally) speed up this routine if you only need static derivatives,
                # or longitudinal derivatives, etc.

            # These lines make a copy of the original operating point, incremented by the finite difference amount
            # along the variable defined by derivative_denominator.
            incremented_op_point = self.op_point.copy()
            if d == "alpha":
                incremented_op_point.alpha += finite_difference_amounts["alpha"]
            elif d == "beta":
                incremented_op_point.beta += finite_difference_amounts["beta"]
            elif d == "p":
                incremented_op_point.p += finite_difference_amounts["p"]
            elif d == "q":
                incremented_op_point.q += finite_difference_amounts["q"]
            elif d == "r":
                incremented_op_point.r += finite_difference_amounts["r"]
            else:
                raise ValueError(f"Invalid value of d: {d}!")

            aerobuildup_incremented = self.copy()
            aerobuildup_incremented.op_point = incremented_op_point
            run_incremented = aerobuildup_incremented.run()

            for derivative_numerator in [
                "CL",
                "CD",
                "CY",
                "Cl",
                "Cm",
                "Cn",
            ]:
                derivative_name = derivative_numerator + abbreviations[d]  # Gives "CLa"
                run_base[derivative_name] = (
                        (  # Finite-difference out the derivatives
                                run_incremented[derivative_numerator] - run_base[derivative_numerator]
                        ) / finite_difference_amounts[d]
                        * scaling_factors[d]
                )

            ### Try to compute and append neutral point, if possible
            if d == "alpha":
                Cma = run_base["Cma"]
                CLa = np.where(
                    run_base["CLa"] == 0,
                    np.nan,
                    run_base["CLa"],
                )

                run_base["x_np"] = self.xyz_ref[0] - (Cma / CLa * self.airplane.c_ref)

            if d == "beta":
                Cnb = run_base["Cnb"]
                CYb = np.where(
                    run_base["CYb"] == 0,
                    np.nan,
                    run_base["CYb"],
                )

                run_base["x_np_lateral"] = self.xyz_ref[0] - (Cnb / CYb * self.airplane.b_ref)

        return run_base

    def wing_aerodynamics(self) -> AeroComponentResults:

        if self.verbose:
            print("Meshing...")

        ##### Make Panels
        front_left_vertices = []
        back_left_vertices = []
        back_right_vertices = []
        front_right_vertices = []
        airfoils: List[Airfoil] = []
        control_surfaces: List[List[ControlSurface]] = []

        for wing in self.airplane.wings:  # subdivide the wing in more spanwise sections
            if self.spanwise_resolution > 1:
                wing = wing.subdivide_sections(
                    ratio=self.spanwise_resolution,
                    spacing_function=self.spanwise_spacing_function
                )

            points, faces = wing.mesh_thin_surface(
                method="quad",
                chordwise_resolution=1,
                add_camber=False,
            )

            front_left_vertices.append(points[faces[:, 0], :])
            back_left_vertices.append(points[faces[:, 1], :])
            back_right_vertices.append(points[faces[:, 2], :])
            front_right_vertices.append(points[faces[:, 3], :])

            wing_airfoils = []
            wing_control_surfaces = []

            for xsec_a, xsec_b in zip(wing.xsecs[:-1], wing.xsecs[1:]):  # Do the right side
                wing_airfoils.append(
                    xsec_a.airfoil.blend_with_another_airfoil(
                        airfoil=xsec_b.airfoil,
                        blend_fraction=0.5,
                    )
                )
                wing_control_surfaces.append(xsec_a.control_surfaces)

            airfoils.extend(wing_airfoils)
            control_surfaces.extend(wing_control_surfaces)

            if wing.symmetric:  # Do the left side, if applicable
                airfoils.extend(wing_airfoils)

                def mirror_control_surface(surf: ControlSurface) -> ControlSurface:
                    if surf.symmetric:
                        return surf
                    else:
                        surf = surf.copy()
                        surf.deflection *= -1
                        return surf

                symmetric_wing_control_surfaces = [
                    [mirror_control_surface(surf) for surf in surfs]
                    for surfs in wing_control_surfaces
                ]

                control_surfaces.extend(symmetric_wing_control_surfaces)

        front_left_vertices = np.concatenate(front_left_vertices)
        back_left_vertices = np.concatenate(back_left_vertices)
        back_right_vertices = np.concatenate(back_right_vertices)
        front_right_vertices = np.concatenate(front_right_vertices)

        ### Compute panel statistics
        diag1 = front_right_vertices - back_left_vertices
        diag2 = front_left_vertices - back_right_vertices
        cross = np.cross(diag1, diag2)
        cross_norm = np.linalg.norm(cross, axis=1)
        normal_directions = cross / tall(cross_norm)
        areas = cross_norm / 2

        # Compute the location of points of interest on each panel
        left_vortex_vertices = 0.75 * front_left_vertices + 0.25 * back_left_vertices
        right_vortex_vertices = 0.75 * front_right_vertices + 0.25 * back_right_vertices
        vortex_centers = (left_vortex_vertices + right_vortex_vertices) / 2
        vortex_bound_leg = right_vortex_vertices - left_vortex_vertices
        vortex_bound_leg_norm = np.linalg.norm(vortex_bound_leg, axis=1)
        chord_vectors = (
                (back_left_vertices + back_right_vertices) / 2 -
                (front_left_vertices + front_right_vertices) / 2
        )
        chords = np.linalg.norm(chord_vectors, axis=1)
        wing_directions = vortex_bound_leg / tall(vortex_bound_leg_norm)
        local_forward_direction = np.cross(normal_directions, wing_directions)

        ### Save things to the instance for later access
        self.front_left_vertices = front_left_vertices
        self.back_left_vertices = back_left_vertices
        self.back_right_vertices = back_right_vertices
        self.front_right_vertices = front_right_vertices
        self.airfoils: List[Airfoil] = airfoils
        self.control_surfaces: List[List[ControlSurface]] = control_surfaces
        self.normal_directions = normal_directions
        self.areas = areas
        self.left_vortex_vertices = left_vortex_vertices
        self.right_vortex_vertices = right_vortex_vertices
        self.vortex_centers = vortex_centers
        self.vortex_bound_leg = vortex_bound_leg
        self.chord_vectors = chord_vectors
        self.chords = chords
        self.local_forward_direction = local_forward_direction
        self.n_panels = areas.shape[0]

        ##### Setup Operating Point
        if self.verbose:
            print("Calculating the freestream influence...")
        steady_freestream_velocity = self.op_point.compute_freestream_velocity_geometry_axes()  # Direction the wind is GOING TO, in geometry axes coordinates
        steady_freestream_direction = steady_freestream_velocity / np.linalg.norm(steady_freestream_velocity)

        steady_freestream_velocities = np.tile(
            wide(steady_freestream_velocity),
            reps=(self.n_panels, 1)
        )
        steady_freestream_directions = np.tile(
            wide(steady_freestream_direction),
            reps=(self.n_panels, 1)
        )

        rotation_freestream_velocities = self.op_point.compute_rotation_velocity_geometry_axes(
            points=vortex_centers
        )

        freestream_velocities = steady_freestream_velocities + rotation_freestream_velocities
        # Nx3, represents the freestream velocity at each panel collocation point (c)

        # freestream_influences = np.sum(freestream_velocities * normal_directions, axis=1)

        ### Save things to the instance for later access
        self.steady_freestream_velocity = steady_freestream_velocity
        self.steady_freestream_direction = steady_freestream_direction
        self.freestream_velocities = freestream_velocities

        ##### Compute the linearization quantities (CL0 and CLa) of each airfoil section
        alpha_geometrics = 90 - np.arccosd(
            np.sum(
                steady_freestream_directions * normal_directions,
                axis=1
            )
        )

        cos_sweeps = np.sum(
            steady_freestream_directions * -local_forward_direction,
            axis=1
        )

        machs = self.op_point.mach() * cos_sweeps

        Res = (
                      self.op_point.velocity *
                      chords /
                      self.op_point.atmosphere.kinematic_viscosity()
              ) * cos_sweeps

        # ### Do a central finite-difference in alpha to obtain CL0 and CLa quantities
        # finite_difference_alpha_amount = 1  # degree
        #
        # aero_finite_differences = [
        #     af.get_aero_from_neuralfoil(
        #         alpha=alpha_geometrics[i] + finite_difference_alpha_amount * np.array([-1, 1]),
        #         Re=Res[i],
        #         mach=machs[i],
        #         control_surfaces=control_surfaces[i]
        #     )
        #     for i, af in enumerate(airfoils)
        # ]
        #
        # CLs_at_alpha_geometric = [
        #     np.mean(aero["CL"])
        #     for aero in aero_finite_differences
        # ]
        # CLas = [
        #     np.diff(aero["CL"])[0] / (2 * np.radians(finite_difference_alpha_amount))
        #     for aero in aero_finite_differences
        # ]
        #
        # # Regularize CLas to always be positive and not-too-close-to-zero
        # CLas = [
        #     np.softmax(CLa, 1, softness=0.5)
        #     for CLa in CLas
        # ]

        ### OVERWRITE CL pre-calculation with more theory-based CLa (improves Cma prediction accuracy due to FD error)
        CLs_at_alpha_geometric = [
            af.get_aero_from_neuralfoil(
                alpha=alpha_geometrics[i],
                Re=Res[i],
                mach=machs[i],
                control_surfaces=control_surfaces[i],
                model_size=self.model_size,
            )["CL"]
            for i, af in enumerate(airfoils)
        ]

        CLas = 2 * np.pi * np.ones(len(CLs_at_alpha_geometric))

        ##### Setup Geometry
        ### Calculate AIC matrix
        if self.verbose:
            print("Calculating the collocation influence matrix...")

        u_centers_unit, v_centers_unit, w_centers_unit = calculate_induced_velocity_horseshoe(
            x_field=tall(vortex_centers[:, 0]),
            y_field=tall(vortex_centers[:, 1]),
            z_field=tall(vortex_centers[:, 2]),
            x_left=wide(left_vortex_vertices[:, 0]),
            y_left=wide(left_vortex_vertices[:, 1]),
            z_left=wide(left_vortex_vertices[:, 2]),
            x_right=wide(right_vortex_vertices[:, 0]),
            y_right=wide(right_vortex_vertices[:, 1]),
            z_right=wide(right_vortex_vertices[:, 2]),
            trailing_vortex_direction=(
                steady_freestream_direction
                if self.align_trailing_vortices_with_wind else
                np.array([1, 0, 0])
            ),
            gamma=1.,
            vortex_core_radius=self.vortex_core_radius
        )

        AIC = (
                u_centers_unit * tall(normal_directions[:, 0]) +
                v_centers_unit * tall(normal_directions[:, 1]) +
                w_centers_unit * tall(normal_directions[:, 2])
        )
        # Influence of panel j's vortex strength [m^2/s] on panel i's normal-flow-velocity [m/s]

        alpha_influence_matrix = AIC / self.op_point.velocity
        # Influence of panel j's vortex strength [m^2/s] on panel i's alpha [radians]

        ##### Calculate Vortex Strengths
        if self.verbose:
            print("Calculating vortex center strengths (assembling and solving linear system)...")

        V_freestream_cross_li = np.cross(steady_freestream_velocities, self.vortex_bound_leg, axis=1)
        V_freestream_cross_li_magnitudes = np.linalg.norm(V_freestream_cross_li, axis=1)

        velocity_magnitude_perpendiculars = self.op_point.velocity * cos_sweeps

        A = alpha_influence_matrix * np.tile(wide(CLas), (self.n_panels, 1)) - np.diag(
            2 * V_freestream_cross_li_magnitudes / velocity_magnitude_perpendiculars ** 2 / areas
        )

        b = -1 * np.array(CLs_at_alpha_geometric)

        vortex_strengths = np.linalg.solve(A, b)
        self.vortex_strengths = vortex_strengths

        ##### Evaluate the aerodynamics with induced effects
        alpha_induced = np.degrees(alpha_influence_matrix @ vortex_strengths)

        alphas = tall(alpha_geometrics) + tall(alpha_induced)

        aeros = [
            af.get_aero_from_neuralfoil(
                alpha=alphas[i],
                Re=Res[i],
                mach=machs[i],
                control_surfaces=control_surfaces[i],
                model_size=self.model_size,
            )
            for i, af in enumerate(airfoils)
        ]
        CLs = np.array([aero["CL"] for aero in aeros])
        CDs = np.array([aero["CD"] for aero in aeros])
        CMs = np.array([aero["CM"] for aero in aeros])

        ##### Calculate forces
        ### Calculate Near-Field Forces and Moments
        # Governing Equation: The force on a straight, small vortex filament is F = rho * cross(V, l) * gamma,
        # where rho is density, V is the velocity vector, cross() is the cross product operator,
        # l is the vector of the filament itself, and gamma is the circulation.

        if self.verbose:
            print("Calculating induced forces on each panel...")
        # Calculate the induced velocity at the center of each bound leg

        velocities = self.get_velocity_at_points(
            points=vortex_centers,
            vortex_strengths=vortex_strengths
        )  # fuse added here

        velocity_magnitudes = np.linalg.norm(velocities, axis=1)

        # Calculate forces_inviscid_geometry, the force on the ith panel. Note that this is in GEOMETRY AXES,
        # not WIND AXES or BODY AXES.
        Vi_cross_li = np.cross(velocities, vortex_bound_leg, axis=1)

        forces_inviscid_geometry = self.op_point.atmosphere.density() * Vi_cross_li * tall(self.vortex_strengths)
        moments_inviscid_geometry = np.cross(
            np.add(vortex_centers, -wide(np.array(self.xyz_ref))),
            forces_inviscid_geometry
        )

        # Calculate total forces and moments
        force_inviscid_geometry = np.sum(forces_inviscid_geometry, axis=0)
        moment_inviscid_geometry = np.sum(moments_inviscid_geometry, axis=0)

        if self.verbose:
            print("Calculating profile forces and moments...")
        forces_profile_geometry = (
                0.5 * self.op_point.atmosphere.density() * velocities * tall(velocity_magnitudes)
                * tall(CDs) * tall(areas)
        )

        moments_profile_geometry = np.cross(
            np.add(vortex_centers, -wide(np.array(self.xyz_ref))),
            forces_profile_geometry
        )
        force_profile_geometry = np.sum(forces_profile_geometry, axis=0)
        moment_profile_geometry = np.sum(moments_profile_geometry, axis=0)

        # Compute pitching moment

        bound_leg_YZ = vortex_bound_leg
        bound_leg_YZ[:, 0] = 0
        moments_pitching_geometry = (0.5 * self.op_point.atmosphere.density() * tall(velocity_magnitudes ** 2)) \
                                    * tall(CMs) * tall(chords ** 2) * bound_leg_YZ
        moment_pitching_geometry = np.sum(moments_pitching_geometry, axis=0)

        if self.verbose:
            print("Calculating total forces and moments...")

        force_total_geometry = np.add(force_inviscid_geometry, force_profile_geometry)
        moment_total_geometry = np.add(moment_inviscid_geometry, moment_profile_geometry) + moment_pitching_geometry

        return self.AeroComponentResults(
            s_ref=self.airplane.s_ref,
            c_ref=self.airplane.c_ref,
            b_ref=self.airplane.b_ref,
            op_point=self.op_point,
            F_g=force_total_geometry,
            M_g=moment_total_geometry,
        )

    def get_induced_velocity_at_points(self,
                                       points: np.ndarray,
                                       vortex_strengths: np.ndarray = None
                                       ) -> np.ndarray:
        """
        Computes the induced velocity at a set of points in the flowfield.

        Args:
            points: A Nx3 array of points that you would like to know the induced velocities at. Given in geometry axes.

        Returns: A Nx3 of the induced velocity at those points. Given in geometry axes.

        """
        if vortex_strengths is None:
            try:
                vortex_strengths = self.vortex_strengths
            except AttributeError:
                raise ValueError(
                    "`LiftingLine.vortex_strengths` doesn't exist, so you need to pass in the `vortex_strengths` parameter.")

        u_induced, v_induced, w_induced = calculate_induced_velocity_horseshoe(
            x_field=tall(points[:, 0]),
            y_field=tall(points[:, 1]),
            z_field=tall(points[:, 2]),
            x_left=wide(self.left_vortex_vertices[:, 0]),
            y_left=wide(self.left_vortex_vertices[:, 1]),
            z_left=wide(self.left_vortex_vertices[:, 2]),
            x_right=wide(self.right_vortex_vertices[:, 0]),
            y_right=wide(self.right_vortex_vertices[:, 1]),
            z_right=wide(self.right_vortex_vertices[:, 2]),
            trailing_vortex_direction=(
                self.steady_freestream_direction
                if self.align_trailing_vortices_with_wind else
                np.array([1, 0, 0])
            ),
            gamma=wide(vortex_strengths),
            vortex_core_radius=self.vortex_core_radius
        )
        u_induced = np.sum(u_induced, axis=1)
        v_induced = np.sum(v_induced, axis=1)
        w_induced = np.sum(w_induced, axis=1)

        V_induced = np.stack([
            u_induced, v_induced, w_induced
        ], axis=1)

        return V_induced

    def get_velocity_at_points(self,
                               points: np.ndarray,
                               vortex_strengths: np.ndarray = None,
                               ) -> np.ndarray:
        """
        Computes the velocity at a set of points in the flowfield.

        Args:
            points: A Nx3 array of points that you would like to know the velocities at. Given in geometry axes.

        Returns: A Nx3 of the velocity at those points. Given in geometry axes.

        """
        V_induced = self.get_induced_velocity_at_points(
            points=points,
            vortex_strengths=vortex_strengths,
        )

        rotation_freestream_velocities = np.array(self.op_point.compute_rotation_velocity_geometry_axes(
            points
        ))

        freestream_velocities = np.add(wide(self.steady_freestream_velocity), rotation_freestream_velocities)

        V = V_induced + freestream_velocities

        # if self.airplane.fuselages:
        #     V_induced_fuselage = self.calculate_fuselage_influences(
        #         points=points
        #     )
        #     V = V + V_induced_fuselage

        return V

    def calculate_fuselage_influences(self, points: np.ndarray) -> np.ndarray:

        this_fuse_centerline_points = []  # fuselage sections centres
        this_fuse_radii = []

        for fuse in self.airplane.fuselages:  # iterating through the airplane fuselages
            for xsec_num in range(len(fuse.xsecs)):  # iterating through the current fuselage sections
                this_fuse_xsec = fuse.xsecs[xsec_num]
                this_fuse_centerline_points.append(this_fuse_xsec.xyz_c)
                this_fuse_radii.append(this_fuse_xsec.width / 2)

        this_fuse_centerline_points = np.stack(
            this_fuse_centerline_points,
            axis=0
        )
        this_fuse_centerline_points = (this_fuse_centerline_points[1:, :] +
                                       this_fuse_centerline_points[:-1, :]) / 2
        this_fuse_radii = np.array(this_fuse_radii)

        areas = np.pi * this_fuse_radii ** 2
        freestream_x_component = self.op_point.compute_freestream_velocity_geometry_axes()[
            0]  # TODO add in rotation corrections, add in doublets for alpha
        sigmas = freestream_x_component * np.diff(areas)

        u_induced_fuse, v_induced_fuse, w_induced_fuse = calculate_induced_velocity_point_source(
            x_field=tall(points[:, 0]),
            y_field=tall(points[:, 1]),
            z_field=tall(points[:, 2]),
            x_source=wide(this_fuse_centerline_points[:, 0]),
            y_source=wide(this_fuse_centerline_points[:, 1]),
            z_source=wide(this_fuse_centerline_points[:, 2]),
            sigma=wide(sigmas),
            viscous_radius=0.0001,
        )

        # # Compressibility
        # dy *= self.beta
        # dz *= self.beta

        # For now, we're just putting a point source at the middle... # TODO make an actual line source
        # source_x = (dx[:, 1:] + dx[:, :-1]) / 2
        # source_y = (dy[:, 1:] + dy[:, :-1]) / 2
        # source_z = (dz[:, 1:] + dz[:, :-1]) / 2

        fuselage_influences_x = np.sum(u_induced_fuse, axis=1)
        fuselage_influences_y = np.sum(v_induced_fuse, axis=1)
        fuselage_influences_z = np.sum(w_induced_fuse, axis=1)

        fuselage_influences = np.stack([
            fuselage_influences_x, fuselage_influences_y, fuselage_influences_z
        ], axis=1)

        return fuselage_influences

    def calculate_streamlines(self,
                              seed_points: np.ndarray = None,
                              n_steps: int = 300,
                              length: float = None
                              ) -> np.ndarray:
        """
        Computes streamlines, starting at specific seed points.

        After running this function, a new instance variable `VortexLatticeFilaments.streamlines` is computed

        Uses simple forward-Euler integration with a fixed spatial stepsize (i.e., velocity vectors are normalized
        before ODE integration). After investigation, it's not worth doing fancier ODE integration methods (adaptive
        schemes, RK substepping, etc.), due to the near-singular conditions near vortex filaments.

        Args:

            seed_points: A Nx3 ndarray that contains a list of points where streamlines are started. Will be
            auto-calculated if not specified.

            n_steps: The number of individual streamline steps to trace. Minimum of 2.

            length: The approximate total length of the streamlines desired, in meters. Will be auto-calculated if
            not specified.

        Returns:
            streamlines: a 3D array with dimensions: (n_seed_points) x (3) x (n_steps).
            Consists of streamlines data.

            Result is also saved as an instance variable, VortexLatticeMethod.streamlines.

        """
        if self.verbose:
            print("Calculating streamlines...")
        if length is None:
            length = self.airplane.c_ref * 5
        if seed_points is None:
            left_TE_vertices = self.back_left_vertices
            right_TE_vertices = self.back_right_vertices
            N_streamlines_target = 200
            seed_points_per_panel = np.maximum(1, N_streamlines_target // len(left_TE_vertices))

            nondim_node_locations = np.linspace(0, 1, seed_points_per_panel + 1)
            nondim_seed_locations = (nondim_node_locations[1:] + nondim_node_locations[:-1]) / 2

            seed_points = np.concatenate([
                x * left_TE_vertices + (1 - x) * right_TE_vertices
                for x in nondim_seed_locations
            ])

        streamlines = np.empty((len(seed_points), 3, n_steps))
        streamlines[:, :, 0] = seed_points
        for i in range(1, n_steps):
            V = self.get_velocity_at_points(streamlines[:, :, i - 1])
            streamlines[:, :, i] = (
                    streamlines[:, :, i - 1] +
                    length / n_steps * V / tall(np.linalg.norm(V, axis=1))
            )

        self.streamlines = streamlines

        if self.verbose:
            print("Streamlines calculated.")

        return streamlines

    def draw(self,
             c: np.ndarray = None,
             cmap: str = None,
             colorbar_label: str = None,
             show: bool = True,
             show_kwargs: Dict = None,
             draw_streamlines=True,
             recalculate_streamlines=False,
             backend: str = "pyvista"
             ):
        """
        Draws the solution. Note: Must be called on a SOLVED AeroProblem object.
        To solve an AeroProblem, use opti.solve(). To substitute a solved solution, use ap = sol(ap).
        :return:
        """
        if show_kwargs is None:
            show_kwargs = {}

        if c is None:
            c = self.vortex_strengths
            colorbar_label = "Vortex Strengths"

        if draw_streamlines:
            if (not hasattr(self, 'streamlines')) or recalculate_streamlines:
                self.calculate_streamlines()

        if backend == "plotly":
            from aerosandbox.visualization.plotly_Figure3D import Figure3D
            fig = Figure3D()

            for i in range(len(self.front_left_vertices)):
                fig.add_quad(
                    points=[
                        self.front_left_vertices[i, :],
                        self.back_left_vertices[i, :],
                        self.back_right_vertices[i, :],
                        self.front_right_vertices[i, :],
                    ],
                    intensity=c[i],
                    outline=True,
                )

            if draw_streamlines:
                for i in range(self.streamlines.shape[0]):
                    fig.add_streamline(self.streamlines[i, :, :].T)

            return fig.draw(
                show=show,
                colorbar_title=colorbar_label
                               ** show_kwargs,
            )

        elif backend == "pyvista":
            import pyvista as pv
            plotter = pv.Plotter()
            plotter.title = "ASB LiftingLine"
            plotter.add_axes()
            plotter.show_grid(color='gray')

            ### Draw the airplane mesh
            points = np.concatenate([
                self.front_left_vertices,
                self.back_left_vertices,
                self.back_right_vertices,
                self.front_right_vertices
            ])
            N = len(self.front_left_vertices)
            range_N = np.arange(N)
            faces = tall(range_N) + wide(np.array([0, 1, 2, 3]) * N)

            mesh = pv.PolyData(
                *mesh_utils.convert_mesh_to_polydata_format(points, faces)
            )
            scalar_bar_args = {}
            if colorbar_label is not None:
                scalar_bar_args["title"] = colorbar_label
            plotter.add_mesh(
                mesh=mesh,
                scalars=c,
                show_edges=True,
                show_scalar_bar=c is not None,
                scalar_bar_args=scalar_bar_args,
                cmap=cmap,
            )

            ### Draw the streamlines
            if draw_streamlines:
                import aerosandbox.tools.pretty_plots as p
                for i in range(self.streamlines.shape[0]):
                    plotter.add_mesh(
                        pv.Spline(self.streamlines[i, :, :].T),
                        color=p.adjust_lightness("#7700FF", 1.5),
                        opacity=0.7,
                        line_width=1
                    )

            if show:
                plotter.show(**show_kwargs)
            return plotter

        else:
            raise ValueError("Bad value of `backend`!")
            # # Fuselages
            # for fuse_id in range(len(self.airplane.fuselages)):
            #     fuse = self.airplane.fuselages[fuse_id]  # type: Fuselage
            #
            #     for xsec_id in range(len(fuse.xsecs) - 1):
            #         xsec_1 = fuse.xsecs[xsec_id]  # type: FuselageXSec
            #         xsec_2 = fuse.xsecs[xsec_id + 1]  # type: FuselageXSec
            #
            #         r1 = xsec_1.equivalent_radius(preserve="area")
            #         r2 = xsec_2.equivalent_radius(preserve="area")
            #         points_1 = np.zeros((fuse.xsec_perimeter, 3))
            #         points_2 = np.zeros((fuse.xsec_perimeter, 3))
            #         for point_index in range(fuse.xsec_perimeter):
            #             from aerosandbox.numpy import rotation_matrix_3D
            #             rot = rotation_matrix_3D(
            #                 2 * np.pi * point_index / fuse.xsec_perimeter,
            #                 [1, 0, 0],
            #                 True
            #             ).toarray()
            #             points_1[point_index, :] = rot @ np.array([0, 0, r1])
            #             points_2[point_index, :] = rot @ np.array([0, 0, r2])
            #         points_1 = points_1 + np.array(xsec_1.xyz_c).reshape(-1)
            #         points_2 = points_2 + np.array(xsec_2.xyz_c).reshape(-1)
            #
            #         for point_index in range(fuse.circumferential_panels):
            #
            #             fig.add_quad(points=[
            #                 points_1[(point_index) % fuse.xsec_perimeter, :],
            #                 points_1[(point_index + 1) % fuse.xsec_perimeter, :],
            #                 points_2[(point_index + 1) % fuse.xsec_perimeter, :],
            #                 points_2[(point_index) % fuse.xsec_perimeter, :],
            #             ],
            #                 intensity=0,
            #             )

            # if draw_streamlines:
            #     if (not hasattr(self, 'streamlines')) or recalculate_streamlines:
            #         if self.verbose:
            #             print("Calculating streamlines...")
            #         seed_points = (back_left_vertices + back_right_vertices) / 2
            #         self.calculate_streamlines(seed_points=seed_points)
            #
            #     if self.verbose:
            #         print("Parsing streamline data...")
            #     n_streamlines = self.streamlines[0].shape[0]
            #     n_timesteps = len(self.streamlines)
            #
            #     for streamlines_num in range(n_streamlines):
            #         streamline = [self.streamlines[ts][streamlines_num, :] for ts in range(n_timesteps)]
            #         fig.add_streamline(
            #             points=streamline,
            #             mirror=self.run_symmetric
            #         )


if __name__ == '__main__':
    import aerosandbox as asb
    import aerosandbox.numpy as np
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    airplane = asb.Airplane(
        wings=[
            asb.Wing(
                symmetric=False,
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[0, 0, 0],
                        chord=1,
                        airfoil=asb.Airfoil("naca0012")
                    ),
                    asb.WingXSec(
                        xyz_le=[0, 5, 0],
                        chord=1,
                        airfoil=asb.Airfoil("naca0012")
                    ),
                ]
            )
        ]
    )

    ll = LiftingLine(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=100,
            alpha=5,
            beta=0,
        ),
        xyz_ref=[0.25, 0, 0],
        spanwise_resolution=5,
    )

    aero = ll.run()

    print(aero["CL"])
    print(aero["CL"] / aero["CD"])


    @np.vectorize
    def get_aero(resolution):
        return LiftingLine(
            airplane=airplane,
            op_point=OperatingPoint(
                velocity=100,
                alpha=5,
                beta=0,
            ),
            xyz_ref=[0.25, 0, 0],
            spanwise_resolution=resolution,
        ).run()

    resolutions = 1 + np.arange(20)
    aeros = get_aero(resolutions)

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(3, 1)
    ax[0].semilogx(
        resolutions,
        [aero["CL"] for aero in aeros]
    )
    ax[0].plot(
        resolutions,
        np.array([aero["CL"] for aero in aeros]) * (resolutions) / (resolutions + 0.3),
        # np.array([aero["CL"] for aero in aeros]) * (resolutions) / (resolutions + 0.3),
    )
    ax[1].plot(
        resolutions,
        [aero["CD"] for aero in aeros]
    )
    ax[1].plot(
        resolutions,
        np.array([aero["CD"] for aero in aeros]) * (resolutions + 0.2) / (resolutions),
    )
    ax[2].plot(
        resolutions,
        [aero["CL"] / aero["CD"] for aero in aeros]
    )
    p.show_plot()



import aerosandbox.numpy as np
from aerosandbox.geometry.airfoil.airfoil import Airfoil
from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_parameters
from aerosandbox.modeling.splines.hermite import linear_hermite_patch, cubic_hermite_patch, cosine_hermite_patch
from typing import Union, Dict, List
import warnings


class KulfanAirfoil(Airfoil):
    def __init__(self,
                 name: str = "Untitled",
                 lower_weights: np.ndarray = None,
                 upper_weights: np.ndarray = None,
                 leading_edge_weight: float = 0.,
                 TE_thickness: float = 0.,
                 N1: float = 0.5,
                 N2: float = 1.0,
                 ):
        ### Handle the airfoil name
        self.name = name

        ### Check to see if the airfoil is a "known" airfoil, based on its name.
        if (
                lower_weights is None and
                upper_weights is None
        ):  # Try to fall back on parameters from the coordinate airfoil, if it's something from the UIUC database

            class IgnoreUserWarnings:  # A context manager to ignore UserWarnings
                def __enter__(self):
                    warnings.filterwarnings("ignore", category=UserWarning)

                def __exit__(self, exc_type, exc_val, exc_tb):
                    warnings.resetwarnings()

            with IgnoreUserWarnings():
                coordinate_airfoil = Airfoil(name)

            if coordinate_airfoil.coordinates is None:
                raise ValueError("You must either:\n"
                                 "\t* Specify both `lower_weights` and `upper_weights`, at minimum"
                                 "\t* Give an airfoil `name` corresponding to an airfoil in the UIUC database, or a NACA airfoil.")

            else:
                parameters = get_kulfan_parameters(
                    coordinates=coordinate_airfoil.coordinates,
                    n_weights_per_side=8,
                    N1=N1,
                    N2=N2,
                    normalize_coordinates=True,
                    use_leading_edge_modification=True,
                )

                lower_weights = parameters["lower_weights"]
                upper_weights = parameters["upper_weights"]
                leading_edge_weight = parameters["leading_edge_weight"]
                TE_thickness = parameters["TE_thickness"]

        ### Handle the Kulfan parameters
        self.lower_weights = lower_weights
        self.upper_weights = upper_weights
        self.leading_edge_weight = leading_edge_weight
        self.TE_thickness = TE_thickness
        self.N1 = N1
        self.N2 = N2

    def __repr__(self) -> str:
        return f"Airfoil {self.name} (Kulfan / CST parameterization)"

    @property
    def kulfan_parameters(self):
        return {
            "lower_weights"      : self.lower_weights,
            "upper_weights"      : self.upper_weights,
            "leading_edge_weight": self.leading_edge_weight,
            "TE_thickness"       : self.TE_thickness,
        }

    @property
    def coordinates(self) -> np.ndarray:
        return self.to_airfoil().coordinates

    @coordinates.setter
    def coordinates(self, value):
        raise TypeError("The coordinates of a `KulfanAirfoil` can't be modified directly, "
                        "as they're a function of the Kulfan parameters.\n"
                        "Instead, you can either modify the Kulfan parameters directly, or use the "
                        "more general (coordinate-parameterized) `asb.Airfoil` class.")

    def to_airfoil(self,
                   n_coordinates_per_side=200,
                   spacing_function_per_side=np.cosspace,
                   ) -> Airfoil:
        x_upper = spacing_function_per_side(1, 0, n_coordinates_per_side)[:-1]
        upper = self.upper_coordinates(x_over_c=x_upper)

        x_lower = spacing_function_per_side(0, 1, n_coordinates_per_side)
        lower = self.lower_coordinates(x_over_c=x_lower)

        return Airfoil(
            name=self.name,
            coordinates=np.concatenate([
                upper,
                lower
            ], axis=0)
        )

    def repanel(self,
                n_points_per_side: int = 100,
                spacing_function_per_side=np.cosspace,
                ) -> 'Airfoil':
        return self.to_airfoil(
            n_coordinates_per_side=n_points_per_side,
            spacing_function_per_side=spacing_function_per_side
        )

    def normalize(
            self,
            return_dict: bool = False,
    ) -> Union['KulfanAirfoil', Dict[str, Union['KulfanAirfoil', float]]]:
        """
        Returns a copy of the Airfoil with a new set of `coordinates`, such that:
            - The leading edge (LE) is at (0, 0)
            - The trailing edge (TE) is at (1, 0)
            - The chord length is equal to 1

        The trailing-edge (TE) point is defined as the midpoint of the line segment connecting the first and last coordinate points (upper and lower surface TE points, respectively). The TE point is not necessarily one of the original points in the airfoil coordinates (`Airfoil.coordinates`); in general, it will not be one of the points if the TE thickness is nonzero.

        The leading-edge (LE) point is defined as the coordinate point with the largest Euclidian distance from the trailing edge. (In other words, if you were to center a circle on the trailing edge and progressively grow it, what's the last coordinate point that it would intersect?) The LE point is always one of the original points in the airfoil coordinates.

        The chord is defined as the Euclidian distance between the LE and TE points.

        Coordinate modifications to achieve the constraints described above (LE @ origin, TE at (1, 0), and chord of 1) are done by means of a translation and rotation.

        Args:

            return_dict: Determines the output type of the function.
                - If `False` (default), returns a copy of the Airfoil with the new coordinates.
                - If `True`, returns a dictionary with keys:

                        - "airfoil": a copy of the Airfoil with the new coordinates

                        - "x_translation": the amount by which the airfoil's LE was translated in the x-direction

                        - "y_translation": the amount by which the airfoil's LE was translated in the y-direction

                        - "scale_factor": the amount by which the airfoil was scaled (if >1, the airfoil had to get
                            bigger)

                        - "rotation_angle": the angle (in degrees) by which the airfoil was rotated about the LE.
                            Sign convention is that positive angles rotate the airfoil counter-clockwise.

                    All of thes values represent the "required change", e.g.:

                        - "x_translation" is the amount by which the airfoil's LE had to be translated in the
                            x-direction to get it to the origin.

                        - "rotation_angle" is the angle (in degrees) by which the airfoil had to be rotated (CCW).

        Returns: Depending on the value of `return_dict`, either:

            - A copy of the airfoil with the new coordinates (default), or

            - A dictionary with keys "airfoil", "x_translation", "y_translation", "scale_factor", and "rotation_angle".
                documentation for `return_tuple` for more information.
        """
        ### A KulfanAirfoil is already normalized by definition, so we essentially just return the same airfoil.
        # (This method serves to overwrite the Airfoil.normalize() method, which is not applicable to KulfanAirfoils.)

        if not return_dict:
            return self
        else:
            return {
                "airfoil"       : self,
                "x_translation" : 0,
                "y_translation" : 0,
                "scale_factor"  : 1,
                "rotation_angle": 0
            }

    def draw(self,
             *args,
             draw_markers=False,
             **kwargs
             ):
        return self.to_airfoil().draw(
            *args,
            draw_markers=draw_markers,
            **kwargs
        )

    def get_aero_from_neuralfoil(self,
                                 alpha: Union[float, np.ndarray],
                                 Re: Union[float, np.ndarray],
                                 mach: Union[float, np.ndarray] = 0.,
                                 n_crit: Union[float, np.ndarray] = 9.0,
                                 xtr_upper: Union[float, np.ndarray] = 1.0,
                                 xtr_lower: Union[float, np.ndarray] = 1.0,
                                 model_size: str = "large",
                                 control_surfaces: List["ControlSurface"] = None,
                                 include_360_deg_effects: bool = True,
                                 ) -> Dict[str, Union[float, np.ndarray]]:
        ### Validate inputs
        if (
                (np.length(self.lower_weights) != 8) or
                (np.length(self.upper_weights) != 8)
        ):
            raise NotImplementedError("NeuralFoil is only trained to handle exactly 8 CST coefficients per side.")

        if (
                self.N1 != 0.5 or
                self.N2 != 1.0
        ):
            raise NotImplementedError("NeuralFoil is only trained to handle airfoils with N1 = 0.5 and N2 = 1.0.")

        ### Set up inputs
        if control_surfaces is None:
            control_surfaces = []

        alpha = np.mod(alpha + 180, 360) - 180  # Enforce periodicity of alpha

        ##### Evaluate the control surfaces of the airfoil
        effective_d_alpha = 0.
        effective_CD_multiplier_from_control_surfaces = 1.

        for surf in control_surfaces:

            effectiveness = 1 - np.maximum(0, surf.hinge_point + 1e-16) ** 2.751428551177291
            # From XFoil-based study at `/AeroSandbox/studies/ControlSurfaceEffectiveness/`

            effective_d_alpha += surf.deflection * effectiveness

            effective_CD_multiplier_from_control_surfaces *= (
                    2 + (surf.deflection / 11.5) ** 2 - (1 + (surf.deflection / 11.5) ** 2) ** 0.5
            )
            # From fit to wind tunnel data from Hoerner, "Fluid Dynamic Drag", 1965. Page 13-13, Figure 32,
            # "Variation of section drag coefficient of a horizontal tail surface at constant C_L"

        ##### Use NeuralFoil to evaluate the incompressible aerodynamics of the airfoil
        import neuralfoil as nf
        nf_aero = nf.get_aero_from_kulfan_parameters(
            kulfan_parameters=dict(
                lower_weights=self.lower_weights,
                upper_weights=self.upper_weights,
                leading_edge_weight=self.leading_edge_weight,
                TE_thickness=self.TE_thickness,
            ),
            alpha=alpha + effective_d_alpha,
            Re=Re,
            n_crit=n_crit,
            xtr_upper=xtr_upper,
            xtr_lower=xtr_lower,
            model_size=model_size
        )

        CL = nf_aero["CL"]
        CD = nf_aero["CD"] * effective_CD_multiplier_from_control_surfaces
        CM = nf_aero["CM"]
        # Cpmin_0 = nf_aero["Cpmin"]
        Cpmin_0 = np.softmin(
            *[
                1 - nf_aero[f"upper_bl_ue/vinf_{i}"] ** 2
                for i in range(len(nf.bl_x_points))
            ],
            *[
                1 - nf_aero[f"lower_bl_ue/vinf_{i}"] ** 2
                for i in range(len(nf.bl_x_points))
            ],
            softness=0.01
        )
        Top_Xtr = nf_aero["Top_Xtr"]
        Bot_Xtr = nf_aero["Bot_Xtr"]

        ##### Extend aerodynamic data to 360 degrees (post-stall) using wind tunnel behavior here.
        if include_360_deg_effects:
            from aerosandbox.aerodynamics.aero_2D.airfoil_polar_functions import airfoil_coefficients_post_stall

            CL_if_separated, CD_if_separated, CM_if_separated = airfoil_coefficients_post_stall(
                airfoil=self,
                alpha=alpha
            )
            import aerosandbox.library.aerodynamics as lib_aero

            # These values are set relatively high because NeuralFoil extrapolates quite well past stall
            alpha_stall_positive = 20
            alpha_stall_negative = -20

            # This will be an input to a tanh() sigmoid blend via asb.numpy.blend(), so a value of 1 means the flow is
            # ~90% separated, and a value of -1 means the flow is ~90% attached.
            is_separated = np.softmax(
                alpha - alpha_stall_positive,
                alpha_stall_negative - alpha
            ) / 3

            CL = np.blend(
                is_separated,
                CL_if_separated,
                CL
            )
            CD = np.exp(np.blend(
                is_separated,
                np.log(CD_if_separated + lib_aero.Cf_flat_plate(Re_L=Re, method="turbulent")),
                np.log(CD)
            ))
            CM = np.blend(
                is_separated,
                CM_if_separated,
                CM
            )
            """

            Separated Cpmin_0 model is a very rough fit to Figure 3 of:

            Shademan & Naghib-Lahouti, "Effects of aspect ratio and inclination angle on aerodynamic loads of a flat 
            plate", Advances in Aerodynamics. 
            https://www.researchgate.net/publication/342316140_Effects_of_aspect_ratio_and_inclination_angle_on_aerodynamic_loads_of_a_flat_plate

            """
            Cpmin_0 = np.blend(
                is_separated,
                -1 - 0.5 * np.sind(alpha) ** 2,
                Cpmin_0
            )

            Top_Xtr = np.blend(
                is_separated,
                0.5 - 0.5 * np.tanh(10 * np.sind(alpha)),
                Top_Xtr
            )
            Bot_Xtr = np.blend(
                is_separated,
                0.5 + 0.5 * np.tanh(10 * np.sind(alpha)),
                Bot_Xtr
            )

        ###### Add compressibility effects

        ### Step 1: compute mach_crit, the critical Mach number
        """
        Below is a function that computes the critical Mach number from the incompressible Cp_min.

        It's based on a Laitone-rule compressibility correction (similar to Prandtl-Glauert or Karman-Tsien, 
        but higher order), together with the Cp_sonic relation. When the Laitone-rule Cp equals Cp_sonic, we have reached
        the critical Mach number.

        This approach does not admit explicit solution for the Cp0 -> M_crit relation, so we instead regress a 
        relationship out using symbolic regression. In effect, this is a curve fit to synthetic data.

        See fits at: /AeroSandbox/studies/MachFitting/CriticalMach/
        """
        Cpmin_0 = np.softmin(
            Cpmin_0,
            0,
            softness=0.001
        )

        mach_crit = (
                            1.011571026701678
                            - Cpmin_0
                            + 0.6582431351007195 * (-Cpmin_0) ** 0.6724789439840343
                    ) ** -0.5504677038358711

        mach_dd = mach_crit + (0.1 / 320) ** (1 / 3)  # drag divergence Mach number
        # Relation taken from W.H. Mason's Korn Equation

        ### Step 2: adjust CL, CD, CM, Cpmin by compressibility effects
        gamma = 1.4  # Ratio of specific heats, 1.4 for air (mostly diatomic nitrogen and oxygen)
        beta_squared_ideal = 1 - mach ** 2
        beta = np.softmax(
            beta_squared_ideal,
            -beta_squared_ideal,
            softness=0.5  # Empirically tuned to data
        ) ** 0.5

        CL = CL / beta
        # CD = CD / beta
        CM = CM / beta

        # Prandtl-Glauert
        Cpmin = Cpmin_0 / beta

        # Karman-Tsien
        # Cpmin = Cpmin_0 / (
        #     beta
        #     + mach ** 2 / (1 + beta) * (Cpmin_0 / 2)
        # )

        # Laitone's rule
        # Cpmin = Cpmin_0 / (
        #         beta
        #         + (mach ** 2) * (1 + (gamma - 1) / 2 * mach ** 2) / (1 + beta) * (Cpmin_0 / 2)
        # )

        ### Step 3: modify CL based on buffet and supersonic considerations
        # Accounts approximately for the lift drop due to buffet.
        buffet_factor = np.blend(
            50 * (mach - (mach_dd + 0.04)),  # Tuned to RANS CFD data empirically
            np.blend(
                (mach - 1) / 0.1,
                1,
                0.5
            ),
            1,
        )

        # Accounts for the fact that theoretical CL_alpha goes from 2 * pi (subsonic) to 4 (supersonic),
        # following linearized supersonic flow on a thin airfoil.
        cla_supersonic_ratio_factor = np.blend(
            (mach - 1) / 0.1,
            4 / (2 * np.pi),
            1,
        )
        CL = CL * buffet_factor * cla_supersonic_ratio_factor

        # Step 4: Account for wave drag
        t_over_c = self.max_thickness()

        CD_wave = np.where(
            mach < mach_crit,
            0,
            np.where(
                mach < mach_dd,
                80 * (mach - mach_crit) ** 4,
                np.where(
                    mach < 1.1,
                    cosine_hermite_patch(
                        mach,
                        x_a=mach_dd,
                        x_b=1.1,
                        f_a=80 * (0.1 / 320) ** (4 / 3),
                        f_b=0.8 * t_over_c,
                        dfdx_a=0.1,
                        dfdx_b=-0.8 * t_over_c * 8,
                    ),
                    np.blend(
                        8 * 2 * (mach - 1.1) / (1.2 - 0.8),
                        0.8 * 0.8 * t_over_c,
                        1.2 * 0.8 * t_over_c,
                    )
                )
            )
        )

        CD = CD + CD_wave

        # Step 5: If beyond M_crit or if separated, move the airfoil aerodynamic center back to x/c = 0.5 (Mach tuck)
        has_aerodynamic_center_shift = (mach - (mach_dd + 0.06)) / 0.06

        if include_360_deg_effects:
            has_aerodynamic_center_shift = np.softmax(
                is_separated,
                has_aerodynamic_center_shift,
                softness=0.1
            )

        CM = CM + np.blend(
            has_aerodynamic_center_shift,
            -0.25 * np.cosd(alpha) * CL - 0.25 * np.sind(alpha) * CD,
            0,
        )

        N = len(nf.bl_x_points)

        return {
            "analysis_confidence": nf_aero["analysis_confidence"],
            "CL"       : CL,
            "CD"       : CD,
            "CM"       : CM,
            "Cpmin"    : Cpmin,
            "Top_Xtr"  : Top_Xtr,
            "Bot_Xtr"  : Bot_Xtr,
            "mach_crit": mach_crit,
            "mach_dd"  : mach_dd,
            "Cpmin_0"  : Cpmin_0,
            **{f"upper_bl_theta_{i}": nf_aero[f"upper_bl_theta_{i}"] for i in range(N)},
            **{f"upper_bl_H_{i}": nf_aero[f"upper_bl_H_{i}"] for i in range(N)},
            **{f"upper_bl_ue/vinf_{i}": nf_aero[f"upper_bl_ue/vinf_{i}"] for i in range(N)},
            **{f"lower_bl_theta_{i}": nf_aero[f"lower_bl_theta_{i}"] for i in range(N)},
            **{f"lower_bl_H_{i}": nf_aero[f"lower_bl_H_{i}"] for i in range(N)},
            **{f"lower_bl_ue/vinf_{i}": nf_aero[f"lower_bl_ue/vinf_{i}"] for i in range(N)},
        }

    def upper_coordinates(self,
                          x_over_c: Union[float, np.ndarray] = np.linspace(1, 0, 101),
                          ) -> np.ndarray:
        x_over_c = np.array(x_over_c)

        # Class function
        C = (x_over_c) ** self.N1 * (1 - x_over_c) ** self.N2

        from scipy.special import comb

        def shape_function(w):
            # Shape function (Bernstein polynomials)
            N = np.length(w) - 1  # Order of Bernstein polynomials

            K = comb(N, np.arange(N + 1))  # Bernstein polynomial coefficients

            dims = (np.length(w), np.length(x_over_c))

            def wide(vector):
                return np.tile(np.reshape(vector, (1, dims[1])), (dims[0], 1))

            def tall(vector):
                return np.tile(np.reshape(vector, (dims[0], 1)), (1, dims[1]))

            S_matrix = (
                    tall(K) * wide(x_over_c) ** tall(np.arange(N + 1)) *
                    wide(1 - x_over_c) ** tall(N - np.arange(N + 1))
            )  # Bernstein polynomial coefficients * weight matrix
            S_x = np.sum(tall(w) * S_matrix, axis=0)

            # Calculate y output
            y = C * S_x
            return y

        y_upper = shape_function(self.upper_weights)

        # Add trailing-edge (TE) thickness
        y_upper += x_over_c * self.TE_thickness / 2

        # Add Kulfan's leading-edge-modification (LEM)
        y_upper += self.leading_edge_weight * (x_over_c) * (1 - x_over_c) ** (np.length(self.upper_weights) + 0.5)

        return np.stack((
            np.reshape(x_over_c, (-1)),
            y_upper,
        ), axis=1)

    def lower_coordinates(self,
                          x_over_c: Union[float, np.ndarray] = np.linspace(0, 1, 101),
                          ) -> np.ndarray:
        x_over_c = np.array(x_over_c)

        # Class function
        C = (x_over_c) ** self.N1 * (1 - x_over_c) ** self.N2

        from scipy.special import comb

        def shape_function(w):
            # Shape function (Bernstein polynomials)
            N = np.length(w) - 1  # Order of Bernstein polynomials

            K = comb(N, np.arange(N + 1))  # Bernstein polynomial coefficients

            dims = (np.length(w), np.length(x_over_c))

            def wide(vector):
                return np.tile(np.reshape(vector, (1, dims[1])), (dims[0], 1))

            def tall(vector):
                return np.tile(np.reshape(vector, (dims[0], 1)), (1, dims[1]))

            S_matrix = (
                    tall(K) * wide(x_over_c) ** tall(np.arange(N + 1)) *
                    wide(1 - x_over_c) ** tall(N - np.arange(N + 1))
            )  # Bernstein polynomial coefficients * weight matrix
            S_x = np.sum(tall(w) * S_matrix, axis=0)

            # Calculate y output
            y = C * S_x
            return y

        y_lower = shape_function(self.lower_weights)

        # Add trailing-edge (TE) thickness
        y_lower -= x_over_c * self.TE_thickness / 2

        # Add Kulfan's leading-edge-modification (LEM)
        y_lower += self.leading_edge_weight * (x_over_c) * (1 - x_over_c) ** (np.length(self.lower_weights) + 0.5)

        return np.stack((
            np.reshape(x_over_c, (-1)),
            y_lower
        ), axis=1)

    def local_camber(self,
                     x_over_c: Union[float, np.ndarray] = np.linspace(0, 1, 101),
                     ) -> Union[float, np.ndarray]:
        upper = self.upper_coordinates(x_over_c=x_over_c)
        lower = self.lower_coordinates(x_over_c=x_over_c)

        if np.isscalar(x_over_c):
            return (upper[0, 1] + lower[0, 1]) / 2
        else:
            return (upper[:, 1] + lower[:, 1]) / 2

    def local_thickness(self,
                        x_over_c: Union[float, np.ndarray] = np.linspace(0, 1, 101),
                        ) -> Union[float, np.ndarray]:
        upper = self.upper_coordinates(x_over_c=x_over_c)
        lower = self.lower_coordinates(x_over_c=x_over_c)

        if np.isscalar(x_over_c):
            return (upper[0, 1] - lower[0, 1])
        else:
            return (upper[:, 1] - lower[:, 1])

    def LE_radius(self, relative_softness: float = 0.03):
        # LE_radius_upper = np.where(
        #     self.upper_weights[0] > 0,
        #     self.upper_weights[0] ** 2,
        #     0
        # ) / 2
        # LE_radius_lower = np.where(
        #     self.lower_weights[0] < 0,
        #     self.lower_weights[0] ** 2,
        #     0
        # ) / 2

        return np.softmin_scalefree(
            np.where(
                self.upper_weights[0] > 0,
                self.upper_weights[0],
                0
            ) ** 2,
            np.where(
                self.lower_weights[0] < 0,
                self.lower_weights[0],
                0
            ) ** 2,
            relative_softness=relative_softness
        ) / 2


    def TE_angle(self):
        return np.degrees(
            np.arctan(self.upper_weights[-1]) -
            np.arctan(self.lower_weights[-1]) +
            np.arctan(self.TE_thickness)
        )

    def area(self):

        def get_area_of_side(weights):
            from scipy.special import beta, comb

            N = np.length(weights) - 1
            i = np.arange(N + 1)
            area_of_each_mode = comb(N, i) * beta(
                self.N1 + i + 1,
                self.N2 + N - i + 1
            )
            return np.sum(
                area_of_each_mode * weights
            )

        return (
                get_area_of_side(self.upper_weights) -
                get_area_of_side(self.lower_weights) +
                (self.TE_thickness / 2)
        )

    def set_TE_thickness(self,
                         thickness: float = 0.,
                         ) -> 'KulfanAirfoil':
        """
        Creates a modified copy of the KulfanAirfoil that has a specified trailing-edge thickness.

        Note that the trailing-edge thickness is given nondimensionally (e.g., as a fraction of chord).

        Args:
            thickness: The target trailing-edge thickness, given nondimensionally (e.g., as a fraction of chord).

        Returns: The modified KulfanAirfoil.

        """
        return KulfanAirfoil(
            name=self.name,
            lower_weights=self.lower_weights,
            upper_weights=self.upper_weights,
            leading_edge_weight=self.leading_edge_weight,
            TE_thickness=thickness,
            N1=self.N1,
            N2=self.N2,
        )

    def scale(self,
              scale_x: float = 1.,
              scale_y: float = 1.,
              ) -> "KulfanAirfoil":
        """
        Scales a KulfanAirfoil about the origin.

        Args:

            scale_x: Amount to scale in the x-direction. Note: not supported by KulfanAirfoil due to inherent
                limitations of parameterization; only given here so that argument symmetry to Airfoil.scale() is
                retained. Raises a ValueError if modified, along with instructions to use `Airfoil` if needed.

            scale_y: Amount to scale in the y-direction. Scaling by a negative y-value will result in `lower_weights`
                and `upper_weights` being flipped as appropriate.

        Returns: A copy of the KulfanAirfoil with appropriate scaling applied.
        """
        if scale_x != 1:
            raise ValueError("\n".join([
                "Scaling a KulfanAirfoil in the x-direction is not supported due to inherent limitations of the",
                "Kulfan parameterization. If you need to scale in the x-direction: "
                "\t- Convert to an Airfoil first (`KulfanAirfoil.to_airfoil()`)"
                "\t- Scale it (`Airfoil.scale()`)"
                "\t- Convert back to a KulfanAirfoil (`Airfoil.to_kulfan_airfoil()`)"
            ]))

        if scale_y >= 0:
            return KulfanAirfoil(
                name=self.name,
                lower_weights=self.lower_weights * scale_y,
                upper_weights=self.upper_weights * scale_y,
                leading_edge_weight=self.leading_edge_weight * scale_y,
                TE_thickness=self.TE_thickness * scale_y,
                N1=self.N1,
                N2=self.N2,
            )
        else:
            return KulfanAirfoil(
                name=self.name,
                lower_weights=self.upper_weights * scale_y,
                upper_weights=self.lower_weights * scale_y,
                leading_edge_weight=self.leading_edge_weight * scale_y,
                TE_thickness=self.TE_thickness * (-1 * scale_y),
                N1=self.N1,
                N2=self.N2,
            )

    def blend_with_another_airfoil(self,
                                   airfoil: Union["KulfanAirfoil", Airfoil],
                                   blend_fraction: float = 0.5,
                                   ) -> "KulfanAirfoil":
        if not isinstance(airfoil, KulfanAirfoil):
            try:
                airfoil = airfoil.to_kulfan_airfoil()
            except AttributeError:
                raise TypeError("The `airfoil` argument should be either a `KulfanAirfoil` or an `Airfoil`.\n"
                                f"You gave an object of type \"{type(airfoil)}\".")

        foil_a = self
        foil_b = airfoil
        a_fraction = 1 - blend_fraction
        b_fraction = blend_fraction

        ### Determine parameters for the blended airfoil
        name = f"{a_fraction * 100:.0f}% {self.name}, {b_fraction * 100:.0f}% {airfoil.name}"

        return KulfanAirfoil(
            name=name,
            lower_weights=a_fraction * foil_a.lower_weights + b_fraction * foil_b.lower_weights,
            upper_weights=a_fraction * foil_a.upper_weights + b_fraction * foil_b.upper_weights,
            leading_edge_weight=a_fraction * foil_a.leading_edge_weight + b_fraction * foil_b.leading_edge_weight,
            TE_thickness=a_fraction * foil_a.TE_thickness + b_fraction * foil_b.TE_thickness,
            N1=a_fraction * foil_a.N1 + b_fraction * foil_b.N1,
            N2=a_fraction * foil_a.N2 + b_fraction * foil_b.N2,
        )



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

"""
These tests solve variants of the Rosenbrock problem:

----------
2-dimensional variant:
    Minimize:
        (a-x)**2 + b*(y-x**2)**2
        for a = 1, b = 100.


N-dimensional variant: 
    Minimize:
        sum(100*(x[i+1]-x[i]**2)**2 + (1-x[i])**2) for i in [1, N-1]

"""
# Constants
a = 1
b = 100


def test_2D_rosenbrock():  # 2-dimensional rosenbrock
    opti = asb.Opti()  # set up an optimization environment

    # Define optimization variables
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)

    # Define objective
    f = (a - x) ** 2 + b * (y - x ** 2) ** 2
    opti.minimize(f)

    # Optimize
    sol = opti.solve()

    for i in [x, y]:
        assert sol(i) == pytest.approx(1, abs=1e-4)


def test_2D_rosenbrock_circle_constrained():  # 2-dimensional rosenbrock, constrained to be within the unit circle.
    opti = asb.Opti()  # set up an optimization environment

    # Define optimization variables
    x = opti.variable(init_guess=0)
    y = opti.variable(init_guess=0)

    # Define constraints
    opti.subject_to(x ** 2 + y ** 2 <= 1)

    # Define objective
    f = (a - x) ** 2 + b * (y - x ** 2) ** 2
    opti.minimize(f)

    # Optimize
    sol = opti.solve()

    # Check
    assert sol(x) == pytest.approx(0.7864, abs=1e-4)
    assert sol(y) == pytest.approx(0.6177, abs=1e-4)
    # (Solution also given here: https://www.mathworks.com/help/optim/ug/example-nonlinear-constrained-minimization.html#brg0p3g-2 )


def test_ND_rosenbrock_constrained(N=10):  # N-dimensional rosenbrock
    # Problem is unimodal for N=2, N=3, and N>=8. Bimodal for 4<=N<=7. Global min is always a vector of ones.

    opti = asb.Opti()  # set up an optimization environment

    x = opti.variable(init_guess=1, n_vars=N)  # vector of variables

    objective = 0
    for i in range(N - 1):
        objective += 100 * (x[i + 1] - x[i] ** 2) ** 2 + (1 - x[i]) ** 2

    opti.subject_to(x >= 0)  # Keeps design space unimodal for all N.

    opti.minimize(objective)

    sol = opti.solve()  # solve

    for i in range(N):
        assert sol(x[i]) == pytest.approx(1, abs=1e-4)


def test_2D_rosenbrock_frozen():
    opti = asb.Opti()  # set up an optimization environment

    # Define optimization variables
    x = opti.variable(init_guess=1)
    y = opti.variable(init_guess=0, freeze=True)

    # Define objective
    f = (a - x) ** 2 + b * (y - x ** 2) ** 2
    opti.minimize(f)

    # Optimize
    sol = opti.solve()

    assert sol(x) == pytest.approx(0.161, abs=1e-3)
    assert sol(y) == pytest.approx(0, abs=1e-3)
    assert sol(f) == pytest.approx(0.771, abs=1e-3)


if __name__ == '__main__':
    pytest.main()



import aerosandbox.numpy as np
from typing import List


def softmax_scalefree(x: List[float]) -> float:
    if len(x) == 1:
        return x[0]
    else:
        softness = np.max(np.array(
            [1e-6] + x
        )) * 0.01

        return np.softmax(
            *x,
            softness=softness
        )



import aerosandbox as asb
import aerosandbox.numpy as np
import aerosandbox.tools.units as u
from .raymer_fudge_factors import advanced_composites
from typing import Union


# From Raymer, Aircraft Design: A Conceptual Approach, 5th Ed.
# Section 15.3.2: Cargo/Transport Weights

def mass_wing(
        wing: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the wing for a cargo/transport aircraft, according to Raymer's Aircraft Design: A Conceptual
    Approach.

    Note: Torenbeek's wing mass model is likely more accurate; see `mass_wing()` in `torenbeek_weights.py` (same
    directory).

    Args:
        wing: The wing object.

        design_mass_TOGW: The design take-off gross weight of the entire airplane [kg].

        ultimate_load_factor: Ultimate load factor of the airplane.

        use_advanced_composites: Whether to use advanced composites for the wing. If True, the wing mass is modified
        accordingly.

    Returns:
        Wing mass [kg].
    """
    airfoil_thicknesses = [
        xsec.airfoil.max_thickness()
        for xsec in wing.xsecs
    ]

    airfoil_t_over_c = np.min(airfoil_thicknesses)

    return (
            0.0051 *
            (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.557 *
            (wing.area('planform') / u.foot ** 2) ** 0.649 *
            wing.aspect_ratio() ** 0.5 *
            airfoil_t_over_c ** -0.4 *
            (1 + wing.taper_ratio()) ** 0.1 *
            np.cosd(wing.mean_sweep_angle()) ** -1 *
            (wing.control_surface_area() / u.foot ** 2) ** 0.1 *
            (advanced_composites["wing"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_hstab(
        hstab: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        wing_to_hstab_distance: float,
        fuselage_width_at_hstab_intersection: float,
        aircraft_y_radius_of_gyration: float = None,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the horizontal stabilizer for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        hstab: The horizontal stabilizer object.

        design_mass_TOGW: The design take-off gross weight of the entire airplane [kg].

        ultimate_load_factor: Ultimate load factor of the airplane.

        wing_to_hstab_distance: Distance from the wing's root-quarter-chord-point to the hstab's
        root-quarter-chord-point [m].

        fuselage_width_at_hstab_intersection: Width of the fuselage at the intersection of the wing and hstab [m].

        aircraft_y_radius_of_gyration: Radius of gyration of the aircraft about the y-axis [m]. If None, estimates
        this as `0.3 * wing_to_hstab_distance`.

        use_advanced_composites: Whether to use advanced composites for the hstab. If True, the hstab mass is modified
        accordingly.

    Returns:
        The mass of the horizontal stabilizer [kg].
    """
    if aircraft_y_radius_of_gyration is None:
        aircraft_y_radius_of_gyration = 0.3 * wing_to_hstab_distance

    area = hstab.area()

    ### Determine if the hstab is all-moving or not
    all_moving = True
    for xsec in hstab.xsecs:
        for control_surface in xsec.control_surfaces:
            if (
                    (control_surface.trailing_edge and control_surface.hinge_point > 0) or
                    (not control_surface.trailing_edge and control_surface.hinge_point < 1)
            ):
                all_moving = False
                break

    return (
            0.0379 *
            (1.143 if all_moving else 1) *
            (1 + fuselage_width_at_hstab_intersection / hstab.span()) ** -0.25 *
            (design_mass_TOGW / u.lbm) ** 0.639 *
            ultimate_load_factor ** 0.10 *
            (area / u.foot ** 2) ** 0.75 *
            (wing_to_hstab_distance / u.foot) ** -1 *
            (aircraft_y_radius_of_gyration / u.foot) ** 0.704 *
            np.cosd(hstab.mean_sweep_angle()) ** -1 *
            hstab.aspect_ratio() ** 0.166 *
            (1 + hstab.control_surface_area() / area) ** 0.1 *
            (advanced_composites["tails"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_vstab(
        vstab: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        wing_to_vstab_distance: float,
        is_t_tail: bool = False,
        aircraft_z_radius_of_gyration: float = None,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the vertical stabilizer for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        vstab: The vertical stabilizer object.

        design_mass_TOGW: The design take-off gross weight of the entire airplane [kg].

        ultimate_load_factor: Ultimate load factor of the airplane.

        wing_to_vstab_distance: Distance from the wing's root-quarter-chord-point to the vstab's
        root-quarter-chord-point [m].

        is_t_tail: Whether the airplane is a T-tail or not.

        aircraft_z_radius_of_gyration: The z-radius of gyration of the entire airplane [m]. If None, estimates this
        as `1 * wing_to_vstab_distance`.

        use_advanced_composites: Whether to use advanced composites for the vstab. If True, the vstab mass is modified
        accordingly.

    Returns:
        The mass of the vertical stabilizer [kg].

    """
    airfoil_thicknesses = [
        xsec.airfoil.max_thickness()
        for xsec in vstab.xsecs
    ]

    airfoil_t_over_c = np.min(airfoil_thicknesses)

    if aircraft_z_radius_of_gyration is None:
        aircraft_z_radius_of_gyration = 1 * wing_to_vstab_distance

    return (
            0.0026 *
            (1 + (1 if is_t_tail else 0)) ** 0.225 *
            (design_mass_TOGW / u.lbm) ** 0.556 *
            ultimate_load_factor ** 0.536 *
            (wing_to_vstab_distance / u.foot) ** -0.5 *
            (vstab.area('planform') / u.foot ** 2) ** 0.5 *
            (aircraft_z_radius_of_gyration / u.foot) ** 0.875 *
            np.cosd(vstab.mean_sweep_angle()) ** -1 *
            vstab.aspect_ratio() ** 0.35 *
            airfoil_t_over_c ** -0.5 *
            (advanced_composites["tails"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_fuselage(
        fuselage: asb.Fuselage,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        L_over_D: float,
        main_wing: asb.Wing,
        n_cargo_doors: int = 1,
        has_aft_clamshell_door: bool = False,
        landing_gear_mounted_on_fuselage: bool = False,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the fuselage for a cargo/transport aircraft, according to Raymer's Aircraft Design: A
    Conceptual Approach.

    Args:
        fuselage: The fuselage object.

        design_mass_TOGW: The design take-off gross weight of the entire airplane [kg].

        ultimate_load_factor: Ultimate load factor of the airplane.

        L_over_D: The lift-to-drag ratio of the airplane in cruise.

        main_wing: The main wing object. Can be:

            * An instance of an AeroSandbox wing object (`asb.Wing`)

            * None, if the airplane has no main wing.

        n_cargo_doors: The number of cargo doors on the fuselage.

        has_aft_clamshell_door: Whether or not the fuselage has an aft clamshell door.

        landing_gear_mounted_on_fuselage: Whether or not the landing gear is mounted on the fuselage.

        use_advanced_composites: Whether to use advanced composites for the fuselage. If True, the fuselage mass is
        modified accordingly.

    Returns:
        The mass of the fuselage [kg].
    """
    K_door = (1 + (0.06 * n_cargo_doors)) * (1.12 if has_aft_clamshell_door else 1)

    K_lg = 1.12 if landing_gear_mounted_on_fuselage else 1

    fuselage_structural_length = fuselage.length()

    if main_wing is not None:
        K_ws = (
                0.75 *
                (
                        (1 + 2 * main_wing.taper_ratio()) /
                        (1 + main_wing.taper_ratio())
                ) *
                (
                        main_wing.span() / fuselage_structural_length *
                        np.tand(main_wing.mean_sweep_angle())
                )
        )
    else:
        K_ws = 0

    return (
            0.3280 *
            K_door *
            K_lg *
            (design_mass_TOGW / u.lbm * ultimate_load_factor) ** 0.5 *
            (fuselage_structural_length / u.foot) ** 0.25 *
            (fuselage.area_wetted() / u.foot ** 2) ** 0.302 *
            (1 + K_ws) ** 0.04 *
            L_over_D ** 0.10 * # L/D
            (advanced_composites["fuselage/nacelle"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_main_landing_gear(
        main_gear_length: float,
        landing_speed: float,
        design_mass_TOGW: float,
        is_kneeling: bool = False,
        n_gear: int = 2,
        n_wheels: int = 12,
        n_shock_struts: int = 4,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the main landing gear for a cargo/transport aircraft, according to Raymer's Aircraft Design:
    A Conceptual Approach.

    Args:
        main_gear_length: length of the main landing gear [m].

        landing_speed: landing speed [m/s].

        design_mass_TOGW: The design take-off gross weight of the entire airplane [kg].

        is_kneeling: whether the main landing gear is capable of kneeling.

        n_gear: number of landing gear.

        n_wheels: number of wheels in total on the main landing gear.

        n_shock_struts: number of shock struts.

        use_advanced_composites: Whether to use advanced composites for the landing gear. If True, the landing gear mass
        is modified accordingly.

    Returns:
        mass of the main landing gear [kg].
    """

    K_mp = 1.126 if is_kneeling else 1

    ultimate_landing_load_factor = n_gear * 1.5

    return (
            0.0106 *
            K_mp *  # non-kneeling LG
            (design_mass_TOGW / u.lbm) ** 0.888 *
            ultimate_landing_load_factor ** 0.25 *
            (main_gear_length / u.inch) ** 0.4 *
            n_wheels ** 0.321 *
            n_shock_struts ** -0.5 *
            (landing_speed / u.knot) ** 0.1 *
            (advanced_composites["landing_gear"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_nose_landing_gear(
        nose_gear_length: float,
        design_mass_TOGW: float,
        is_kneeling: bool = False,
        n_gear: int = 1,
        n_wheels: int = 2,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the nose landing gear for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        nose_gear_length: Length of nose landing gear when fully-extended [m].

        design_mass_TOGW: The design take-off gross weight of the entire airplane [kg].

        is_kneeling: Whether the nose landing gear is capable of kneeling.

        n_gear: Number of nose landing gear.

        n_wheels: Number of wheels in total on the nose landing gear.

        use_advanced_composites: Whether to use advanced composites for the landing gear. If True, the landing gear mass
        is modified accordingly.

    Returns:
        Mass of nose landing gear [kg].
    """
    K_np = 1.15 if is_kneeling else 1

    ultimate_landing_load_factor = n_gear * 1.5

    return (
            0.032 *
            K_np *
            (design_mass_TOGW / u.lbm) ** 0.646 *
            ultimate_landing_load_factor ** 0.2 *
            (nose_gear_length / u.inch) ** 0.5 *
            n_wheels ** 0.45 *
            (advanced_composites["landing_gear"] if use_advanced_composites else 1)
    ) * u.lbm


def mass_nacelles(
        nacelle_length: float,
        nacelle_width: float,
        nacelle_height: float,
        ultimate_load_factor: float,
        mass_per_engine: float,
        n_engines: int,
        is_pylon_mounted: bool = False,
        engines_have_propellers: bool = False,
        engines_have_thrust_reversers: bool = False,
        use_advanced_composites: bool = False,
) -> float:
    """
    Computes the mass of the nacelles for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach. Excludes the engine itself and immediate engine peripherals.

    Args:
        nacelle_length: length of the nacelle, front to back [m]

        nacelle_width: width of the nacelle [m]

        nacelle_height: height of the nacelle, top to bottom [m]

        ultimate_load_factor: ultimate load factor of the aircraft

        mass_per_engine: mass of the engine itself [kg]

        n_engines: number of engines

        is_pylon_mounted: whether the engine is pylon-mounted or not

        engines_have_propellers: whether the engines have propellers or not (e.g., a jet)

        engines_have_thrust_reversers: whether the engines have thrust reversers or not

        use_advanced_composites: Whether to use advanced composites for the nacelles. If True, the nacelles mass
        is modified accordingly.

    Returns:
        mass of the nacelles [kg]
    """
    K_ng = 1.017 if is_pylon_mounted else 1

    K_p = 1.4 if engines_have_propellers else 1
    K_tr = 1.18 if engines_have_thrust_reversers else 1

    mass_per_engine_with_contents = np.softmax(
        (2.331 * (mass_per_engine / u.lbm) ** 0.901) * K_p * K_tr * u.lbm,
        mass_per_engine,
        hardness=10 / mass_per_engine
    )

    nacelle_wetted_area = (
            nacelle_length * nacelle_height * 2 +
            nacelle_width * nacelle_height * 2
    )

    return (
            0.6724 *
            K_ng *
            (nacelle_length / u.foot) ** 0.10 *
            (nacelle_width / u.foot) ** 0.294 *
            (ultimate_load_factor) ** 0.119 *
            (mass_per_engine_with_contents / u.lbm) ** 0.611 *
            (n_engines) ** 0.984 *
            (nacelle_wetted_area / u.foot ** 2) ** 0.224 *
            (advanced_composites["fuselage/nacelle"] if use_advanced_composites else 1)
    )


def mass_engine_controls(
        n_engines: int,
        cockpit_to_engine_length: float,
) -> float:
    """
    Computes the mass of the engine controls for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        n_engines: The number of engines in the aircraft.

        cockpit_to_engine_length: The distance from the cockpit to the engine [m].

    Returns:
        The mass of the engine controls [kg].
    """
    return (
            5 * n_engines +
            0.80 * (cockpit_to_engine_length / u.foot) * n_engines
    ) * u.lbm


def mass_starter(
        n_engines: int,
        mass_per_engine: float,
) -> float:
    """
    Computes the mass of the engine starter for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        n_engines: The number of engines in the aircraft.

        mass_per_engine: The mass of the engine [kg].

    Returns:
        The mass of the engine starter [kg].
    """
    return (
            49.19 * (
            mass_per_engine / u.lbm * n_engines
            / 1000
    ) ** 0.541
    ) * u.lbm


def mass_fuel_system(
        fuel_volume: float,
        n_tanks: int,
        fraction_in_integral_tanks: float = 0.5,
) -> float:
    """
    Computes the mass of the fuel system (e.g., tanks, pumps, but not the fuel itself) for a cargo/transport
    aircraft, according to Raymer's Aircraft Design: A Conceptual Approach.

    Args:
        fuel_volume: The volume of fuel in the aircraft [m^3].

        n_tanks: The number of fuel tanks in the aircraft.

        fraction_in_integral_tanks: The fraction of the fuel volume that is in integral tanks, as opposed to
        protected tanks.

    Returns:
        The mass of the fuel system [kg].
    """

    fraction_in_protected_tanks = 1 - fraction_in_integral_tanks
    return (
            2.405 *
            (fuel_volume / u.gallon) ** 0.606 *
            (1 + fraction_in_integral_tanks) ** -1 *
            (1 + fraction_in_protected_tanks) *
            n_tanks ** 0.5
    ) * u.lbm


def mass_flight_controls(
        airplane: asb.Airplane,
        aircraft_Iyy: float,
        fraction_of_mechanical_controls: int = 0,
) -> float:
    """
    Computes the added mass of the flight control surfaces (and any applicable linkages, in the case of mechanical
    controls) for a cargo/transport aircraft, according to Raymer's Aircraft Design: A Conceptual Approach.

    Args:
        airplane: The airplane to calculate the mass of the flight controls for.

        aircraft_Iyy: The moment of inertia of the aircraft about the y-axis.

        fraction_of_mechanical_controls: The fraction of the flight controls that are mechanical, as opposed to
        hydraulic.

    Returns:
        The mass of the flight controls [kg].
    """

    ### Compute how many functions the control surfaces are performing (e.g., aileron, elevator, flap, rudder, etc.)
    N_functions_performed_by_controls = 0
    for wing in airplane.wings:
        N_functions_performed_by_controls += len(wing.get_control_surface_names())

    ### Compute the control surface area
    control_surface_area = 0
    for wing in airplane.wings:
        control_surface_area += wing.control_surface_area()

    return (
            145.9 *
            N_functions_performed_by_controls ** 0.554 *  # number of functions performed by controls
            (1 + fraction_of_mechanical_controls) ** -1 *
            (control_surface_area / u.foot ** 2) ** 0.20 *
            (aircraft_Iyy / (u.lbm * u.foot ** 2) * 1e-6) ** 0.07
    ) * u.lbm


def mass_APU(
        mass_APU_uninstalled: float,
):
    """
    Computes the mass of the auxiliary power unit (APU) for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        mass_APU_uninstalled: The mass of the APU uninstalled [kg].

    Returns:
        The mass of the APU, as installed [kg].
    """
    return 2.2 * mass_APU_uninstalled


def mass_instruments(
        fuselage: asb.Fuselage,
        main_wing: asb.Wing,
        n_engines: int,
        n_crew: Union[int, float],
        engine_is_reciprocating: bool = False,
        engine_is_turboprop: bool = False,
):
    """
    Computes the mass of the flight instruments for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        fuselage: The fuselage of the airplane.

        main_wing: The main wing of the airplane.

        n_engines: The number of engines on the airplane.

        n_crew: The number of crew members on the airplane. Use 0.5 for a UAV.

        engine_is_reciprocating: Whether the engine is reciprocating.

        engine_is_turboprop: Whether the engine is a turboprop.

    Returns:
        The mass of the instruments [kg]
    """
    K_r = 1.133 if engine_is_reciprocating else 1

    K_tp = 0.793 if engine_is_turboprop else 1

    return (
            4.509 *
            K_r *
            K_tp *
            n_crew ** 0.541 *
            n_engines *
            (fuselage.length() / u.foot * main_wing.span() / u.foot) ** 0.5
    ) * u.lbm


def mass_hydraulics(
        airplane: asb.Airplane,
        fuselage: asb.Fuselage,
        main_wing: asb.Wing,
):
    """
    Computes the mass of the hydraulic system for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        airplane: The airplane to calculate the mass of the hydraulic system for.

        fuselage: The fuselage of the airplane.

        main_wing: The main wing of the airplane.

    Returns:
        The mass of the hydraulic system [kg].
    """
    N_functions_performed_by_controls = 0
    for wing in airplane.wings:
        N_functions_performed_by_controls += len(wing.get_control_surface_names())

    return (
            0.2673 *
            N_functions_performed_by_controls *
            (fuselage.length() / u.foot * main_wing.span() / u.foot) ** 0.937
    ) * u.lbm


def mass_electrical(
        system_electrical_power_rating: float,
        electrical_routing_distance: float,
        n_engines: int,
):
    """
    Computes the mass of the electrical system for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:


        system_electrical_power_rating: The total electrical power rating of the aircraft's electrical system [Watts].

            Typical values:
                * Transport airplane: 40,000 - 60,000 W
                * Fighter/bomber airplane: 110,000 - 160,000 W

        electrical_routing_distance: The electrical routing distance, generators to avionics to cockpit. [meters]

    Returns:

        The mass of the electrical system [kg].

    """

    return (
            7.291 *
            (system_electrical_power_rating / 1e3) ** 0.782 *
            (electrical_routing_distance / u.foot) ** 0.346 *
            (n_engines) ** 0.10
    ) * u.lbm


def mass_avionics(
        mass_uninstalled_avionics: float,
):
    """
    Computes the mass of the avionics for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        mass_uninstalled_avionics: The mass of the avionics, before installation [kg].

    Returns:
        The mass of the avionics, as installed [kg].
    """
    return (
            1.73 *
            (mass_uninstalled_avionics / u.lbm) ** 0.983
    ) * u.lbm


def mass_furnishings(
        n_crew: Union[int, float],
        mass_cargo: float,
        fuselage: asb.Fuselage,
):
    """
    Computes the mass of the furnishings for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach. Does not include cargo handling gear or seats.

    Args:
        n_crew: The number of crew members on the airplane. Use 0.5 for a UAV.

        mass_cargo: The mass of the cargo [kg].

        fuselage: The fuselage of the airplane.

    Returns:
        The mass of the furnishings [kg].
    """
    return (
            0.0577 *
            n_crew ** 0.1 *
            (mass_cargo / u.lbm) ** 0.393 *
            (fuselage.area_wetted() / u.foot ** 2) ** 0.75
    ) * u.lbm


def mass_air_conditioning(
        n_crew: int,
        n_pax: int,
        volume_pressurized: float,
        mass_uninstalled_avionics: float,
):
    """
    Computes the mass of the air conditioning system for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        n_crew: The number of crew members on the airplane.

        n_pax: The number of passengers on the airplane.

        volume_pressurized: The volume of the pressurized cabin [meters^3].

        mass_uninstalled_avionics: The mass of the avionics, before installation [kg].

    Returns:
        The mass of the air conditioning system [kg].
    """
    return (
            62.36 *
            (n_crew + n_pax) ** 0.25 *
            (volume_pressurized / u.foot ** 3 / 1e3) ** 0.604 *
            (mass_uninstalled_avionics / u.lbm) ** 0.10
    ) * u.lbm


def mass_anti_ice(
        design_mass_TOGW: float,
):
    """
    Computes the mass of the anti-ice system for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        design_mass_TOGW: The design takeoff gross weight of the entire airplane [kg].

    Returns:
        The mass of the anti-ice system [kg].
    """
    return 0.002 * design_mass_TOGW


def mass_handling_gear(
        design_mass_TOGW: float,
):
    """
    Computes the mass of the handling gear for a cargo/transport aircraft, according to Raymer's Aircraft
    Design: A Conceptual Approach.

    Args:
        design_mass_TOGW: The design takeoff gross weight of the entire airplane [kg].

    Returns:
        The mass of the handling gear [kg].
    """
    return 3e-4 * design_mass_TOGW


def mass_military_cargo_handling_system(
        cargo_floor_area: float,
):
    """
    Computes the mass of the military cargo handling system for a cargo/transport aircraft, according to Raymer's
    Aircraft Design: A Conceptual Approach.

    Args:
        cargo_floor_area: The floor area of the cargo compartment [meters^2].

    Returns:
        The mass of the military cargo handling system [kg].
    """
    return (
            2.4 *
            (cargo_floor_area / u.foot ** 2)
    ) * u.lbm
